/*****************************************************************************
*                  QLOGIC LINUX SOFTWARE
*
* QLogic  QLA4000 iSCSI driver
* Copyright (C) 2002 Qlogic Corporation 
* (www.qlogic.com)
*
* This program is free software; you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the
* Free Software Foundation; either version 2, or (at your option) any
* later version.
*
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
*
******************************************************************************/

#include <linux/config.h>
#include <linux/module.h>
#include <stdarg.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/segment.h>
#include <asm/byteorder.h>
#include <linux/spinlock.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/pci.h>
#include <linux/proc_fs.h>
#include <linux/blk.h>
#include <linux/tqueue.h>
#include <asm/uaccess.h>

#include <../drivers/scsi/scsi.h>
#include <../drivers/scsi/hosts.h>

#include "qlisioct.h"
#include "qlinioct.h"
#include "qla4x_fw.h"
#include "qla4xxx.h"
#include "ql4print.h"

// KRH: (BEGIN) Define these locally, for now
/*
 * Sub codes for Get Data.
 * Use in combination with INT_GET_DATA as the ioctl code
 */
#define INT_SC_GET_DRIVER_DEBUG_LEVEL       2

/*
 * Sub codes for Set Data.
 * Use in combination with INT_SET_DATA as the ioctl code
 */
#define INT_SC_SET_DRIVER_DEBUG_LEVEL	2

/*
 * Sub codes for Reset
 * Use in combination with INT_CC_HBA_RESET as the ioctl code
 */
#define INT_SC_HBA_RESET			0
#define INT_SC_FIRMWARE_RESET		1
#define INT_SC_TARGET_WARM_RESET		2
#define INT_SC_LUN_RESET			3
#define INT_SC_LINUX_HOST_RESET		4
#define INT_SC_LINUX_BUS_RESET		5
#define INT_SC_LINUX_DEVICE_RESET		6
//KRH: (END)

/* Defines for Passthru */		    
#define IOCTL_INVALID_STATUS		0xffff
#define IOCTL_PASSTHRU_TOV			30

/* Defines for VendorSpecificStatus */
#define VS_STATUS_SCSI_STATUS_INDEX        	1
#define VS_STATUS_SCSI_COMPLETION_INDEX     2
#define VS_STATUS_SENSE_BYTE_COUNT_INDEX	3
#define VS_STATUS_HOST_STATUS_INDEX		4
#define VS_STATUS_ISCSI_RESPONSE_INDEX	5
#define VS_STATUS_STATE_FLAGS_INDEX		6
		    
/**************************
 * Global Data
 **************************/
EXT_IOCTL_ISCSI qla4ioctl;


extern uint8_t qla4000_recover_adapter(scsi_qla_host_t *ha, uint8_t ddb_list_todo);

/**************************
 * Functions 
 **************************/

static int qla4extioctl_query_hba_iscsi_node(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int                status = QLA_ERROR;
	static uint32_t           mbox_cmd[MBOX_REG_COUNT];
	static uint32_t           mbox_sts[MBOX_REG_COUNT];
	dma_buf_t          *dma_buf = &ha->dma_buf;
	EXT_HBA_ISCSI_NODE *hba_node;
	INIT_FW_CTRL_BLK   *init_fw_cb;

	ENTER("qla4extioctl_query_hba_iscsi_node");
	spin_lock(&dma_buf->spinlock);

	hba_node = kmalloc(sizeof(*hba_node), GFP_ATOMIC);
	if (!hba_node || !dma_buf->virt_addr || !dma_buf->phys_addr || !ioctl->ResponseAdr)
	{
		printk("qla4extioctl_query_hba_iscsi_node: memory allocation problem\n");
		ioctl->Status = EXT_STATUS_NO_MEMORY;
		goto exit_query_hba_node;
	}

	if (ioctl->ResponseLen < sizeof(*hba_node) || 
	    dma_buf->buf_len < sizeof(*init_fw_cb))
	{
		printk("qla4extioctl_query_hba_iscsi_node: memory area too small\n");
		ioctl->Status = EXT_STATUS_BUFFER_TOO_SMALL;
		goto exit_query_hba_node;
	}

	/*
	 * Send mailbox command
	 */
	memset(&mbox_cmd, 0, sizeof(mbox_cmd));
	mbox_cmd[0] = MBOX_CMD_GET_INIT_FW_CTRL_BLOCK;
	mbox_cmd[2] = LSL(dma_buf->phys_addr);
	mbox_cmd[3] = MSL(dma_buf->phys_addr);

	if (qla4000_mailbox_command(ha, 4, 1, &mbox_cmd[0], &mbox_sts[0]) == QLA_ERROR)
	{
		QL4PRINT(QLP3|QLP4, printk("qla4extioctl_query_hba_iscsi_node: command failed \n"));
		ioctl->Status = EXT_STATUS_ERR;
		ioctl->DetailStatus = mbox_sts[0];
		goto exit_query_hba_node;
	}

	/*
	 * Transfer data from Fw's DEVDB_ENTRY buffer to IOCTL's EXT_HBA_ISCSI_NODE buffer
	 */
	init_fw_cb = (INIT_FW_CTRL_BLK *) dma_buf->virt_addr;

	memset(hba_node, 0, sizeof(*hba_node));
	hba_node->PortNumber = init_fw_cb->PortNumber;
	hba_node->NodeInfo.PortalCount = 1;

	if (init_fw_cb->IPOptions & DDB_IPOPT_IPV6_ADDRESS_ENABLE)
		hba_node->NodeInfo.IPAddr.Type = EXT_DEF_TYPE_ISCSI_IPV6;
	memcpy(hba_node->NodeInfo.IPAddr.IPAddress, init_fw_cb->IPAddr, sizeof(hba_node->NodeInfo.IPAddr.IPAddress));
	memcpy(hba_node->NodeInfo.iSCSIName,        init_fw_cb->iSCSINameString,  sizeof(hba_node->NodeInfo.iSCSIName));
	memcpy(hba_node->NodeInfo.Alias,            init_fw_cb->Alias,  sizeof(hba_node->NodeInfo.Alias));
	
	sprintf(hba_node->DeviceName, "/proc/scsi/%s/HbaApiNode", DRIVER_NAME);


	/*
	 * Copy the IOCTL EXT_HBA_ISCSI_NODE buffer to the user's data space
	 */
	if (copy_to_user((void *)LSL(ioctl->ResponseAdr), hba_node, ioctl->ResponseLen) != 0)
	{
		QL4PRINT(QLP1, printk("qla4extioctl_query_hba_iscsi_node: copy failed\n"));
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_query_hba_node;
	}

	ioctl->Status = EXT_STATUS_OK;
	status = QLA_SUCCESS;

exit_query_hba_node:
	if (hba_node) kfree(hba_node);
	spin_unlock(&dma_buf->spinlock);
	LEAVE("qla4extioctl_query_hba_iscsi_node");
	return status;
}

static int qla4extioctl_query_hba_iscsi_portal(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = QLA_ERROR;
	static uint32_t mbox_cmd[MBOX_REG_COUNT];
	static uint32_t mbox_sts[MBOX_REG_COUNT];
	EXT_HBA_ISCSI_PORTAL *hba_portal;
	FLASH_SYS_INFO *sys_info;
	uint32_t num_valid_ddb_entries;

	ENTER("qla4extioctl_query_hba_iscsi_portal");

	hba_portal = kmalloc(sizeof(*hba_portal), GFP_ATOMIC);
	if (!hba_portal || !ioctl->ResponseAdr)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_query_hba_iscsi_portal: memory allocation problem\n"));
		ioctl->Status = EXT_STATUS_NO_MEMORY;
		goto exit_query_hba_portal;
	}

	if (ioctl->ResponseLen < sizeof(*hba_portal))
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_query_hba_iscsi_portal: memory area too small\n"));
		ioctl->Status = EXT_STATUS_BUFFER_TOO_SMALL;
		goto exit_query_hba_portal;
	}

	/*
	 * Fill in EXT_HBA_ISCSI_PORTAL buffer
	 */
	memset(hba_portal, 0, sizeof(*hba_portal));

	strcpy(hba_portal->DriverVersion, QLA4XXX_DRIVER_VERSION);
	sprintf(hba_portal->FWVersion, "%02d.%02d Patch %02d Build %02d",
		ha->firmware_version[0], ha->firmware_version[1], 
		ha->patch_number, ha->build_number);

	/* ----- Get firmware state information ---- */
	mbox_cmd[0] = MBOX_CMD_GET_FW_STATE;
	if (qla4000_mailbox_command(ha, 1, 4, &mbox_cmd[0], &mbox_sts[0]) != QLA_SUCCESS)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_query_hba_iscsi_portal: MBOX_CMD_GET_FW_STATE failed w/ status %04X\n", mbox_sts[0])); 
		goto exit_query_hba_portal;
	}

	switch (mbox_sts[1])
	{
	case FW_STATE_READY:
		hba_portal->State = EXT_DEF_CARD_STATE_READY;
		break;
	case FW_STATE_CONFIG_WAIT:
		hba_portal->State = EXT_DEF_CARD_STATE_CONFIG_WAIT;
		break;
	case FW_STATE_WAIT_LOGIN:
		hba_portal->State = EXT_DEF_CARD_STATE_LOGIN;
		break;
	case FW_STATE_ERROR:
		hba_portal->State = EXT_DEF_CARD_STATE_ERROR;
		break;
	}

	switch (mbox_sts[3])
	{
	case MEDIA_TYPE_COPPER:
		hba_portal->Type = EXT_DEF_TYPE_COPPER;
		break;
	case MEDIA_TYPE_OPTICAL:
		hba_portal->Type = EXT_DEF_TYPE_OPTICAL;
		break;
	}

	/* ----- Get ddb entry information ---- */
	if (qla4000_get_ddb_entry(ha, 0, NULL, &num_valid_ddb_entries, NULL, NULL) == QLA_ERROR)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_query_hba_iscsi_portal: qla4000_get_ddb_entry!\n")); 
		goto exit_query_hba_portal;
	}
	hba_portal->DiscTargetCount = (uint16_t) num_valid_ddb_entries;

	/* ----- Get flash sys info information ---- */
	spin_lock(&ha->dma_buf.spinlock);
	sys_info = (FLASH_SYS_INFO *) ha->dma_buf.virt_addr;

	mbox_cmd[0] = MBOX_CMD_READ_FLASH;
	mbox_cmd[1] = LSL(ha->dma_buf.phys_addr);
	mbox_cmd[2] = MSL(ha->dma_buf.phys_addr);
	mbox_cmd[3] = FLASH_OFFSET_SYS_INFO;
	mbox_cmd[4] = sizeof(*sys_info);

	if (qla4000_mailbox_command(ha, 5, 1, &mbox_cmd[0], &mbox_sts[0]) != QLA_SUCCESS)
	{
		QL4PRINT(QLP1|QLP4, 
				    printk("qla4extioctl_query_hba_iscsi_portal: MBOX_CMD_READ_FLASH failed w/ status %04X\n", 
				    mbox_sts[0])); 
		spin_unlock(&ha->dma_buf.spinlock);
		goto exit_query_hba_portal;
	}

	hba_portal->SerialNum = sys_info->serialNumber;
	memcpy(hba_portal->IPAddr.IPAddress, ha->ip_address,   		     sizeof(hba_portal->IPAddr.IPAddress));
	memcpy(hba_portal->MacAddr,          sys_info->physAddr[0].address,  sizeof(hba_portal->MacAddr));
	memcpy(hba_portal->Manufacturer,     sys_info->vendorId, 	     sizeof(hba_portal->Manufacturer));
	memcpy(hba_portal->Model,            sys_info->productId,            sizeof(hba_portal->Model));
	spin_unlock(&ha->dma_buf.spinlock);
	
	//memcpy(hba_portal->OptRomVersion, ?, sizeof(hba_portal->OptRomVersion));

	/*
	 * Copy the IOCTL EXT_HBA_ISCSI_PORTAL buffer to the user's data space
	 */
	if (copy_to_user((void *)LSL(ioctl->ResponseAdr), hba_portal, ioctl->ResponseLen) != 0)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_query_hba_iscsi_portal: memory area too small\n"));
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_query_hba_portal;
	}

	ioctl->Status = EXT_STATUS_OK;
	status = QLA_SUCCESS;

exit_query_hba_portal:
	if (hba_portal)	kfree(hba_portal);
	LEAVE("qla4extioctl_query_hba_iscsi_portal");
	return status;
}

static int qla4extioctl_query_disc_iscsi_node(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = QLA_ERROR;
	DEVDB_ENTRY        *fw_ddb_entry = (DEVDB_ENTRY *) ha->dma_buf.virt_addr;
	static EXT_DISC_ISCSI_NODE disc_node;
	ddb_entry_t *ddb_entry;
	
	ENTER("qla4extioctl_query_disc_iscsi_node");
	
	spin_lock(&ha->dma_buf.spinlock);

	if (ioctl->ResponseLen < sizeof(disc_node))
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_query_disc_iscsi_node: response buffer too small\n"));
		ioctl->Status = EXT_STATUS_BUFFER_TOO_SMALL;
		goto exit_disc_node;
	}

	if (ha->dma_buf.buf_len < sizeof(DEVDB_ENTRY))
	{
		if (qla4000_resize_dma_buf(ha, &ha->dma_buf, sizeof(DEVDB_ENTRY)) != QLA_SUCCESS)
		{
			QL4PRINT(QLP1|QLP4, printk("qla4extioctl_query_disc_iscsi_node: unable to allocate memory for dma buffer.\n"));
			ioctl->Status = EXT_STATUS_NO_MEMORY;
			goto exit_disc_node;
		}
	}

	/* ----- get device database entry info from firmware ---- */
	if (qla4000_get_ddb_entry(ha, ioctl->Instance, fw_ddb_entry, NULL, NULL, NULL) != QLA_SUCCESS)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_query_disc_iscsi_node: failed to get DEVDB_ENTRY info.\n"));
		ioctl->Status = EXT_STATUS_ERR;
		ioctl->RequestLen = 0;
		ioctl->DetailStatus = ioctl->Instance;
		goto exit_disc_node;
	}

	/* --- Transfer data from Fw's DEVDB_ENTRY buffer to IOCTL's EXT_DISC_ISCSI_PORTAL buffer --- */
	memset(&disc_node, 0, sizeof(disc_node));
	disc_node.NodeInfo.PortalCount = 1;
	disc_node.NodeInfo.IPAddr.Type = EXT_DEF_TYPE_ISCSI_IP;
	memcpy(disc_node.NodeInfo.IPAddr.IPAddress, fw_ddb_entry->ipAddr, 
	       MIN(sizeof(disc_node.NodeInfo.IPAddr.IPAddress), 
			   sizeof(fw_ddb_entry->ipAddr)));
	strncpy(disc_node.NodeInfo.Alias, fw_ddb_entry->iSCSIAlias, 
		MIN(sizeof(disc_node.NodeInfo.Alias), 
			sizeof(fw_ddb_entry->iSCSIAlias)));
	strncpy(disc_node.NodeInfo.iSCSIName, fw_ddb_entry->iscsiName, 
		MIN(sizeof(disc_node.NodeInfo.iSCSIName), 
			sizeof(fw_ddb_entry->iscsiName)));

	if ((ddb_entry = qla4000_lookup_ddb_by_fw_index(ha, ioctl->Instance)) == NULL)
	{
		QL4PRINT(QLP1|QLP4, 
							printk("qla4extioctl_query_disc_iscsi_node: device handle (0x%x) not logged in. "
							"Dummy target info returned.\n", ioctl->Instance));
		disc_node.SessionID       = 0xDEAD;
		disc_node.ConnectionID    = 0xDEAD;
		disc_node.PortalGroupID   = 0xDEAD;
		disc_node.ScsiAddr.Bus    = 0xFF;
		disc_node.ScsiAddr.Target = 0xFF;
		disc_node.ScsiAddr.Lun    = 0xFF;
	}
	else 
	{
		disc_node.SessionID       = ddb_entry->target_session_id;
		disc_node.ConnectionID    = ddb_entry->connection_id;
		disc_node.PortalGroupID   = 0;
		disc_node.ScsiAddr.Bus    = ddb_entry->bus;
		disc_node.ScsiAddr.Target = ddb_entry->target;
		disc_node.ScsiAddr.Lun    = 0;
	}

	/* --- Copy Results to user space --- */
	if (copy_to_user((void*)LSL(ioctl->ResponseAdr), &disc_node, sizeof(disc_node)) != 0)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_query_disc_iscsi_node: copy error to user space.\n"));
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_disc_node;
	}
	
	ioctl->Status = EXT_STATUS_OK;
    status = QLA_SUCCESS;

exit_disc_node:
	spin_unlock(&ha->dma_buf.spinlock);
	LEAVE("qla4extioctl_query_disc_iscsi_node");
	return status;
}

static int qla4extioctl_query_disc_iscsi_portal(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = QLA_ERROR;
	DEVDB_ENTRY        *fw_ddb_entry = (DEVDB_ENTRY *) ha->dma_buf.virt_addr;
	EXT_DISC_ISCSI_PORTAL disc_portal;

	ENTER("qla4extioctl_query_disc_iscsi_portal");

	spin_lock(&ha->dma_buf.spinlock);

	if (ioctl->ResponseLen < sizeof(disc_portal))
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_query_disc_iscsi_portal: response buffer too small\n"));
		ioctl->Status = EXT_STATUS_BUFFER_TOO_SMALL;
		goto exit_disc_portal;
	}

	if (ha->dma_buf.buf_len < sizeof(DEVDB_ENTRY))
	{
		if (qla4000_resize_dma_buf(ha, &ha->dma_buf, sizeof(DEVDB_ENTRY)) != QLA_SUCCESS)
		{
			QL4PRINT(QLP1|QLP4, printk("qla4extioctl_query_disc_iscsi_portal: unable to allocate memory for dma buffer.\n"));
			ioctl->Status = EXT_STATUS_NO_MEMORY;
			goto exit_disc_portal;
		}
	}

	/* ----- get device database entry info from firmware ---- */
	if (qla4000_get_ddb_entry(ha, ioctl->Instance, fw_ddb_entry, NULL, NULL, NULL) != QLA_SUCCESS)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_query_disc_iscsi_portal: failed to get DEVDB_ENTRY info.\n"));
		ioctl->Status = EXT_STATUS_ERR;
		ioctl->RequestLen = 0;
		ioctl->DetailStatus = ioctl->Instance;
		goto exit_disc_portal;
	}

	/* --- Transfer data from Fw's DEVDB_ENTRY buffer to IOCTL's EXT_DISC_ISCSI_PORTAL buffer --- */
	memset(&disc_portal, 0, sizeof(disc_portal));
	memcpy(disc_portal.IPAddr.IPAddress, fw_ddb_entry->ipAddr, 
	       MIN(sizeof(disc_portal.IPAddr.IPAddress), sizeof(fw_ddb_entry->ipAddr)));
	disc_portal.PortNumber = fw_ddb_entry->portNumber;
	disc_portal.IPAddr.Type = EXT_DEF_TYPE_ISCSI_IP;

	disc_portal.NodeCount = 0;
	strncpy(disc_portal.HostName, fw_ddb_entry->iscsiName, 
		MIN(sizeof(disc_portal.HostName), sizeof(fw_ddb_entry->iscsiName)));

	/* --- Copy Results to user space --- */
	if (copy_to_user((void*)LSL(ioctl->ResponseAdr), &disc_portal, sizeof(disc_portal)) != 0)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_query_disc_iscsi_portal: copy error to user space.\n"));
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_disc_portal;
	}
	
	ioctl->Status = EXT_STATUS_OK;
    status = QLA_SUCCESS;

exit_disc_portal:
	spin_unlock(&ha->dma_buf.spinlock);
	LEAVE("qla4extioctl_query_disc_iscsi_portal");
	return status;
}

static int qla4extioctl_query_driver(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = QLA_ERROR;
	EXT_DRIVER_INFO dinfo;

	//KRH: Incomplete
	ENTER("qla4extioctl_query_driver");
	
	if (ioctl->ResponseLen < sizeof(dinfo))
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_query_driver: response buffer too small\n"));
		ioctl->Status = EXT_STATUS_BUFFER_TOO_SMALL;
		goto exit_query_driver;
	}
	
	memset(&dinfo, 0, sizeof(dinfo));
	memcpy(dinfo.Version, QLA4XXX_DRIVER_VERSION, sizeof(QLA4XXX_DRIVER_VERSION));
	dinfo.NumOfBus        = EXT_DEF_MAX_HBA;
	dinfo.TargetsPerBus   = EXT_DEF_MAX_TARGET;
	dinfo.LunPerTarget    = EXT_DEF_MAX_LUN;
	dinfo.LunPerTargetOS  = EXT_DEF_MAX_BUS;
	//dinfo.MaxTransferLen  = ?;
	//dinfo.MaxDataSegments = ?;
#if BITS_PER_LONG <= 32
	dinfo.DmaBitAddresses = 0;
#else 
	dinfo.DmaBitAddresses = 1; 
#endif
	dinfo.IoMapType       = 1;  // Memory Mapped I/O
	//dinfo.Attrib          = ?;
	//dinfo.InternalFlags   = ?;

	if (copy_to_user((void*)LSL(ioctl->ResponseAdr), &dinfo, sizeof(dinfo)) != 0)
	{
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_query_driver;
	}
	
	ioctl->Status = EXT_STATUS_OK;
    status = QLA_SUCCESS;

exit_query_driver:

	LEAVE("qla4extioctl_query_driver");
	return status;
}

static int qla4extioctl_query_fw(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	EXT_FW_INFO fw_info;
	int         status = QLA_ERROR;

	ENTER("qla4extioctl_query_fw");

	if (ioctl->ResponseLen < sizeof(fw_info))
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_query_fw: response buffer too small\n"));
		ioctl->Status = EXT_STATUS_BUFFER_TOO_SMALL;
		goto exit_query_fw;
	}

	sprintf(fw_info.Version, 
		"FW Version %d.%d Patch %d Build %d",
	 	ha->firmware_version[0], ha->firmware_version[1],
	 	ha->patch_number, ha->build_number);

	if (copy_to_user((void*)LSL(ioctl->ResponseAdr), &fw_info, sizeof(fw_info)) != 0)
	{
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_query_fw;
	}

	ioctl->Status = EXT_STATUS_OK;
	status = QLA_SUCCESS;

exit_query_fw:
	LEAVE("qla4extioctl_query_fw");
	return status;
}

static int qla4extioctl_query_chip(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = QLA_ERROR;
	static uint32_t mbox_cmd[MBOX_REG_COUNT];
	static uint32_t mbox_sts[MBOX_REG_COUNT];
	EXT_CHIP_INFO chip_info;
	FLASH_SYS_INFO *sys_info;

	ENTER("qla4extioctl_query_chip");

	if (!ioctl->ResponseAdr || ioctl->ResponseLen < sizeof(chip_info))
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_query_chip: response buffer too small\n"));
		ioctl->Status = EXT_STATUS_BUFFER_TOO_SMALL;
		goto exit_query_chip;
	}

	/* Fill in structure */
	memset(&chip_info, 0, sizeof(chip_info));

	/* ----- Get flash sys info information ---- */
	spin_lock(&ha->dma_buf.spinlock);
	sys_info = (FLASH_SYS_INFO *) ha->dma_buf.virt_addr;

	mbox_cmd[0] = MBOX_CMD_READ_FLASH;
	mbox_cmd[1] = LSL(ha->dma_buf.phys_addr);
	mbox_cmd[2] = MSL(ha->dma_buf.phys_addr);
	mbox_cmd[3] = FLASH_OFFSET_SYS_INFO;
	mbox_cmd[4] = sizeof(*sys_info);

	if (qla4000_mailbox_command(ha, 5, 1, &mbox_cmd[0], &mbox_sts[0]) != QLA_SUCCESS)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_query_chip: MBOX_CMD_READ_FLASH failed w/ status %04X\n", mbox_sts[0])); 
		spin_unlock(&ha->dma_buf.spinlock);
		goto exit_query_chip;
	}
	chip_info.VendorId    = sys_info->pciDeviceVendor;
	chip_info.DeviceId    = sys_info->pciDeviceId;
	chip_info.SubVendorId = sys_info->pciSubsysVendor;
	chip_info.SubSystemId = sys_info->pciSubsysId;
	spin_unlock(&ha->dma_buf.spinlock);

	/* ----- Get firmware state information ---- */
	mbox_cmd[0] = MBOX_CMD_GET_FW_STATE;
	if (qla4000_mailbox_command(ha, 1, 4, &mbox_cmd[0], &mbox_sts[0]) != QLA_SUCCESS)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_query_chip: MBOX_CMD_GET_FW_STATE failed w/ status %04X\n", mbox_sts[0])); 
		goto exit_query_chip;
	}

	chip_info.BoardID     = mbox_sts[2];

	/* Copy info to caller */
	if (copy_to_user((void*)LSL(ioctl->ResponseAdr), &chip_info, sizeof(chip_info)) != 0)
	{
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_query_chip;
	}

	ioctl->Status = EXT_STATUS_OK;
	status = QLA_SUCCESS;

exit_query_chip:
	LEAVE("qla4extioctl_query_chip");
	return status;
}

static int qla4extioctl_query(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	switch (ioctl->SubCode)
	{
	case EXT_SC_QUERY_HBA_ISCSI_NODE:
		return qla4extioctl_query_hba_iscsi_node(ha, ioctl);

	case EXT_SC_QUERY_HBA_ISCSI_PORTAL:
		return qla4extioctl_query_hba_iscsi_portal(ha, ioctl);

	case EXT_SC_QUERY_DISC_ISCSI_NODE:
		return qla4extioctl_query_disc_iscsi_node(ha, ioctl);

	case EXT_SC_QUERY_DISC_ISCSI_PORTAL:
		return qla4extioctl_query_disc_iscsi_portal(ha, ioctl);

	case EXT_SC_QUERY_DRIVER:
		return qla4extioctl_query_driver(ha, ioctl);

	case EXT_SC_QUERY_FW:
		return qla4extioctl_query_fw(ha, ioctl);

	case EXT_SC_QUERY_CHIP:
		return qla4extioctl_query_chip(ha, ioctl);

	default:
		QL4PRINT(QLP1|QLP4, printk("qla4000_ioctl: unsupported query sub-command code (%X)\n", ioctl->SubCode));
		ioctl->Status = EXT_STATUS_UNSUPPORTED_SUBCODE;
		return(-EINVAL);
	}
}

static int qla4extioctl_reg_aen(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = QLA_ERROR;
	ioctl->Status = EXT_STATUS_UNSUPPORTED_SUBCODE;
	ENTER("qla4extioctl_reg_aen");
	LEAVE("qla4extioctl_reg_aen");
	return status;
}

static int qla4extioctl_get_aen(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = QLA_ERROR;
	ioctl->Status = EXT_STATUS_UNSUPPORTED_SUBCODE;
	ENTER("qla4extioctl_get_aen");
	LEAVE("qla4extioctl_get_aen");
	return status;
}

static int qla4extioctl_get_statistics_get(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = QLA_ERROR;
	ioctl->Status = EXT_STATUS_UNSUPPORTED_SUBCODE;
	ENTER("qla4extioctl_get_statistics_get");
	LEAVE("qla4extioctl_get_statistics_get");
	return status;
}

static int qla4extioctl_get_statistics_iscsi(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = QLA_ERROR;
	ioctl->Status = EXT_STATUS_UNSUPPORTED_SUBCODE;
	ENTER("qla4extioctl_get_statistics_iscsi");
	LEAVE("qla4extioctl_get_statistics_iscsi");
	return status;
}

static int qla4extioctl_get_device_entry_iscsi(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int                     status = QLA_ERROR;
	static uint32_t          		mbox_cmd[MBOX_REG_COUNT];
	static uint32_t          		mbox_sts[MBOX_REG_COUNT];
	dma_buf_t               *dma_buf = &ha->dma_buf;
	DEVDB_ENTRY             *fw_ddb_entry;
	EXT_DEVICE_ENTRY_ISCSI  *dev_entry;

	ENTER("qla4extioctl_get_device_entry_iscsi");

	dev_entry = kmalloc(sizeof(*dev_entry), GFP_ATOMIC);
	if (!dev_entry || !dma_buf->virt_addr || !dma_buf->phys_addr || !ioctl->ResponseAdr)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_get_device_entry_iscsi: memory allocation problem\n"));
		ioctl->Status = EXT_STATUS_NO_MEMORY;
		goto exit_get_dev_entry;
	}

	if (ioctl->ResponseLen < sizeof(*dev_entry) || 
	    dma_buf->buf_len < sizeof(*fw_ddb_entry))
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_get_device_entry_iscsi: memory area too small\n"));
		ioctl->Status = EXT_STATUS_BUFFER_TOO_SMALL;
		goto exit_get_dev_entry;
	}

	/* 
	 * Make the IOCTL call
	 */
	memset(dev_entry, 0, sizeof(*dev_entry));
	if (ioctl->SubCode == EXT_SC_GET_DEVICE_ENTRY_ISCSI)
            mbox_cmd[0] = MBOX_CMD_GET_DATABASE_ENTRY;
	else 
            mbox_cmd[0] = MBOX_CMD_GET_DATABASE_ENTRY_DEFAULTS;
	mbox_cmd[1] = ioctl->Instance;
	mbox_cmd[2] = LSL(dma_buf->phys_addr);
	mbox_cmd[3] = MSL(dma_buf->phys_addr);

	if (qla4000_mailbox_command(ha, 4, 5, &mbox_cmd[0], &mbox_sts[0]) != QLA_SUCCESS)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_get_device_entry_iscsi: get ddb entry for index %d failed!\n", ioctl->Instance));
		ioctl->Status = EXT_STATUS_MAILBOX;
		ioctl->DetailStatus = mbox_sts[0];
		goto exit_get_dev_entry;
	}

	/*
	 * Transfer data from Fw's DEVDB_ENTRY buffer to IOCTL's EXT_DEVICE_ENTRY_ISCSI buffer
	 */
	fw_ddb_entry = dma_buf->virt_addr;

	dev_entry->NumValid 					= mbox_sts[2];
	dev_entry->NextValid 					= mbox_sts[3];
	dev_entry->DeviceState 					= mbox_sts[4];
	dev_entry->Options 						= fw_ddb_entry->options;
	dev_entry->Control 						= fw_ddb_entry->control;
	dev_entry->TargetSessID                 = fw_ddb_entry->TSID;
	memcpy(dev_entry->InitiatorSessID,  fw_ddb_entry->ISID,   sizeof(fw_ddb_entry->ISID));

	dev_entry->DeviceInfo.DeviceType        = EXT_DEF_ISCSI_REMOTE;
	dev_entry->DeviceInfo.ExeThrottle       = fw_ddb_entry->exeThrottle;
	dev_entry->DeviceInfo.InitMarkerlessInt = fw_ddb_entry->initMarkerlessInt;
	dev_entry->DeviceInfo.RetryCount        = fw_ddb_entry->retryCount;
	dev_entry->DeviceInfo.RetryDelay        = fw_ddb_entry->retryDelay;
	dev_entry->DeviceInfo.iSCSIOptions      = fw_ddb_entry->iSCSIOptions;
	dev_entry->DeviceInfo.TCPOptions        = fw_ddb_entry->TCPOptions;
	dev_entry->DeviceInfo.IPOptions         = fw_ddb_entry->IPOptions;
	dev_entry->DeviceInfo.MaxPDUSize        = fw_ddb_entry->maxPDUSize;
	dev_entry->DeviceInfo.FirstBurstSize	= fw_ddb_entry->firstBurstSize;
	dev_entry->DeviceInfo.LogoutMinTime     = fw_ddb_entry->logoutMinTime;
	dev_entry->DeviceInfo.LogoutMaxTime     = fw_ddb_entry->logoutMaxTime;
	dev_entry->DeviceInfo.MaxOutstandingR2T = fw_ddb_entry->maxOutstndngR2T;
	dev_entry->DeviceInfo.KeepAliveTimeout  = fw_ddb_entry->keepAliveTimeout;
	dev_entry->DeviceInfo.PortNumber        = fw_ddb_entry->portNumber;
	dev_entry->DeviceInfo.MaxBurstSize	    = fw_ddb_entry->maxBurstSize;
	dev_entry->DeviceInfo.TaskMgmtTimeout	= fw_ddb_entry->taskMngmntTimeout;
	dev_entry->EntryInfo.PortalCount 		= mbox_sts[2];
	dev_entry->ExeCount 					= fw_ddb_entry->exeCount;
	dev_entry->DDBLink  					= fw_ddb_entry->ddbLink;

	memcpy(dev_entry->UserID,          fw_ddb_entry->userID,   sizeof(dev_entry->UserID));
	memcpy(dev_entry->Password,        fw_ddb_entry->password, sizeof(dev_entry->Password));

	if (fw_ddb_entry->IPOptions & DDB_IPOPT_IPV6_ADDRESS_ENABLE)
		dev_entry->EntryInfo.IPAddr.Type = EXT_DEF_TYPE_ISCSI_IPV6;
	memcpy(dev_entry->DeviceInfo.TargetAddr, fw_ddb_entry->targetAddr, sizeof(dev_entry->DeviceInfo.TargetAddr));
	memcpy(dev_entry->EntryInfo.IPAddr.IPAddress, fw_ddb_entry->ipAddr,     sizeof(dev_entry->EntryInfo.IPAddr.IPAddress));
	memcpy(dev_entry->EntryInfo.iSCSIName,        fw_ddb_entry->iscsiName,  sizeof(dev_entry->EntryInfo.iSCSIName));
	memcpy(dev_entry->EntryInfo.Alias,            fw_ddb_entry->iSCSIAlias, sizeof(dev_entry->EntryInfo.Alias));
	
	QL4PRINT(QLP10, printk("DEVDB_ENTRY structure:\n"));
	qla4000_dump_bytes(QLP10, fw_ddb_entry, sizeof(*fw_ddb_entry));
	QL4PRINT(QLP10, printk("EXT_DEVICE_ENTRY_ISCSI structure:\n"));
	qla4000_dump_bytes(QLP10, dev_entry, sizeof(*dev_entry));
	
	/*
	 * Copy the IOCTL EXT_DEVICE_ENTRY_ISCSI buffer to the user's data space
	 */
	if (copy_to_user((void *)LSL(ioctl->ResponseAdr), dev_entry, ioctl->ResponseLen) != 0)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_get_device_entry_iscsi: memory area too small\n"));
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_get_dev_entry;
	}

	ioctl->Status = EXT_STATUS_OK;
	status = QLA_SUCCESS;

exit_get_dev_entry:
	if (dev_entry) kfree(dev_entry);
	LEAVE("qla4extioctl_get_device_entry_iscsi");
	return status;
}


static int qla4extioctl_get_init_fw_iscsi(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int               status   = QLA_ERROR;
	static uint32_t          mbox_cmd[MBOX_REG_COUNT];
	static uint32_t          mbox_sts[MBOX_REG_COUNT];
	dma_buf_t         *dma_buf = &ha->dma_buf;
	EXT_INIT_FW_ISCSI *init_fw;
	INIT_FW_CTRL_BLK  *init_fw_cb;

	ENTER("qla4extioctl_get_init_fw_iscsi");
	spin_lock(&dma_buf->spinlock);

	init_fw = kmalloc(sizeof(*init_fw), GFP_ATOMIC);
	if (!dma_buf->virt_addr || !dma_buf->phys_addr || !init_fw || 
	    (dma_buf->buf_len < sizeof(INIT_FW_CTRL_BLK)) ||
	    (ioctl->ResponseLen < sizeof(EXT_INIT_FW_ISCSI))
	   )
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_get_init_fw_iscsi: response buffer too small\n"));
		ioctl->Status = EXT_STATUS_BUFFER_TOO_SMALL;
		goto exit_get_init_fw;
	}

	/*
	 * Send mailbox command
	 */
	switch (ioctl->SubCode)
	{
	case EXT_SC_GET_INIT_FW_ISCSI:          mbox_cmd[0] = MBOX_CMD_GET_INIT_FW_CTRL_BLOCK; break;
	case EXT_SC_GET_INIT_FW_DEFAULTS_ISCSI: mbox_cmd[0] = MBOX_CMD_GET_INIT_FW_CTRL_BLOCK_DEFAULTS; break;
	default: 
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_get_init_fw_iscsi: invalid subcode (0x%04X) speficied\n",
								   ioctl->SubCode));
		ioctl->Status = EXT_STATUS_INVALID_PARAM;
		goto exit_get_init_fw;
	}

	mbox_cmd[1] = 0;
	mbox_cmd[2] = LSL(dma_buf->phys_addr);
	mbox_cmd[3] = MSL(dma_buf->phys_addr);

	if (qla4000_mailbox_command(ha, 4, 1, &mbox_cmd[0], &mbox_sts[0]) == QLA_ERROR)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_get_init_fw_iscsi: command failed \n"));
		ioctl->Status = EXT_STATUS_ERR;
		ioctl->DetailStatus = mbox_sts[0];
		goto exit_get_init_fw;
	}

	/*
	 * Transfer Data from DMA buffer to Local buffer
	 */
	init_fw_cb = (INIT_FW_CTRL_BLK *)dma_buf->virt_addr;
	memset(init_fw, 0, sizeof(*init_fw));

	init_fw->Version         = init_fw_cb->Version;
	init_fw->FWOptions       = init_fw_cb->FwOptions;
	init_fw->AddFWOptions    = init_fw_cb->AddFwOptions;
	init_fw->WakeupThreshold = init_fw_cb->WakeupThreshold;
	memcpy(&init_fw->IPAddr.IPAddress,     &init_fw_cb->IPAddr,          EXT_DEF_IP_ADDR_SIZE);
	memcpy(&init_fw->SubnetMask.IPAddress, &init_fw_cb->SubnetMask,      EXT_DEF_IP_ADDR_SIZE);
	memcpy(&init_fw->Gateway.IPAddress,    &init_fw_cb->GatewayIPAddr,   EXT_DEF_IP_ADDR_SIZE);
	memcpy(&init_fw->DNSConfig.IPAddr.IPAddress, &init_fw_cb->DNSIPAddr, EXT_DEF_IP_ADDR_SIZE);
	memcpy(&init_fw->Alias,                &init_fw_cb->Alias,           EXT_DEF_ISCSI_ALIAS_LEN);
	memcpy(&init_fw->iSCSIName,            &init_fw_cb->iSCSINameString, EXT_DEF_ISCSI_NAME_LEN);

	if (init_fw_cb->IPOptions & IPOPT_IPV6_ADDRESS_ENABLE)
	{
		init_fw->IPAddr.Type = 
		    init_fw->SubnetMask.Type = 
		    init_fw->Gateway.Type = 
		    init_fw->DNSConfig.IPAddr.Type = EXT_DEF_TYPE_ISCSI_IPV6;
	}

	init_fw->DeviceInfo.DeviceType        = EXT_DEF_ISCSI_LOCAL;
	init_fw->DeviceInfo.ExeThrottle       = init_fw_cb->ExecThrottle;
	init_fw->DeviceInfo.InitMarkerlessInt = init_fw_cb->InitMarkerlessInt;
	init_fw->DeviceInfo.RetryCount        = init_fw_cb->RetryCount;
	init_fw->DeviceInfo.RetryDelay        = init_fw_cb->RetryDelay;
	init_fw->DeviceInfo.iSCSIOptions      = init_fw_cb->iSCSIOptions;
	init_fw->DeviceInfo.TCPOptions        = init_fw_cb->TCPOptions;
	init_fw->DeviceInfo.IPOptions         = init_fw_cb->IPOptions;
	init_fw->DeviceInfo.MaxPDUSize        = init_fw_cb->MaxPDUSize;
	init_fw->DeviceInfo.FirstBurstSize    = init_fw_cb->FirstBurstSize;
	init_fw->DeviceInfo.LogoutMinTime     = init_fw_cb->LogOutMinTime;
	init_fw->DeviceInfo.LogoutMaxTime     = init_fw_cb->LogOutMaxTime;
	init_fw->DeviceInfo.LogoutMaxTime     = init_fw_cb->LogOutMaxTime;
	init_fw->DeviceInfo.MaxOutstandingR2T = init_fw_cb->MaxOutStndngR2T;
	init_fw->DeviceInfo.KeepAliveTimeout  = init_fw_cb->KeepAliveTimeout;
	init_fw->DeviceInfo.PortNumber        = init_fw_cb->PortNumber;
	init_fw->DeviceInfo.MaxBurstSize      = init_fw_cb->MaxBurstSize;
	//init_fw->DeviceInfo.TaskMgmtTimeout   = init_fw_cb->T;
	memcpy(&init_fw->DeviceInfo.TargetAddr, &init_fw_cb->TargAddr, EXT_DEF_ISCSI_TADDR_SIZE);

	/*
	 * Copy the local data to the user's buffer
	 */
	if (copy_to_user((void*)LSL(ioctl->ResponseAdr), init_fw, sizeof(*init_fw)) != 0)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_get_init_fw_iscsi: unable to copy data to user's memory area\n"));
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_get_init_fw;
	}

	ioctl->Status = EXT_STATUS_OK;
	status = QLA_SUCCESS;

	QL4PRINT(QLP10, printk("EXT_INIT_FW_ISCSI structure:\n"));
	qla4000_dump_bytes(QLP10, init_fw, sizeof(*init_fw));

exit_get_init_fw:
	if (init_fw) kfree(init_fw);
	spin_unlock(&dma_buf->spinlock);
	LEAVE("qla4extioctl_get_init_fw_iscsi");
	return status;
}

static int qla4extioctl_get_data(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	switch (ioctl->SubCode)
	{
	case EXT_SC_GET_STATISTICS_GEN:
		return qla4extioctl_get_statistics_get(ha, ioctl);

	case EXT_SC_GET_STATISTICS_ISCSI:
		return qla4extioctl_get_statistics_iscsi(ha, ioctl);

	case EXT_SC_GET_DEVICE_ENTRY_ISCSI:
    case EXT_SC_GET_DEVICE_ENTRY_DEFAULTS_ISCSI:
		return qla4extioctl_get_device_entry_iscsi(ha, ioctl);

	case EXT_SC_GET_INIT_FW_ISCSI:
	case EXT_SC_GET_INIT_FW_DEFAULTS_ISCSI:
		return qla4extioctl_get_init_fw_iscsi(ha, ioctl);


	default:
		QL4PRINT(QLP1|QLP4, printk("qla4000_ioctl: unsupported external get data sub-command code (%X)\n", ioctl->SubCode));
		ioctl->Status = EXT_STATUS_UNSUPPORTED_SUBCODE;
		return(-EINVAL);
	}
}

static int qla4extioctl_rst_statistics_gen(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = QLA_ERROR;
	ioctl->Status = EXT_STATUS_UNSUPPORTED_SUBCODE;
	ENTER("qla4extioctl_rst_statistics_gen");
	LEAVE("qla4extioctl_rst_statistics_gen");
	return status;
}

static int qla4extioctl_rst_statistics_iscsi(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = QLA_ERROR;
	ioctl->Status = EXT_STATUS_UNSUPPORTED_SUBCODE;
	ENTER("qla4extioctl_rst_statistics_iscsi");
	LEAVE("qla4extioctl_rst_statistics_iscsi");
	return status;
}

static int qla4extioctl_set_device_entry_iscsi(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int         status = QLA_ERROR;
	dma_buf_t   *dma_buf = &ha->dma_buf;
	DEVDB_ENTRY *fw_ddb_entry;
	EXT_DEVICE_ENTRY_ISCSI *dev_entry;

	ENTER("qla4extioctl_set_device_entry_iscsi");

	dev_entry = kmalloc(sizeof(*dev_entry), GFP_ATOMIC);
	if (!dev_entry || !dma_buf->virt_addr || !dma_buf->phys_addr || !ioctl->RequestAdr)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_set_device_entry_iscsi: memory allocation problem\n"));
		ioctl->Status = EXT_STATUS_NO_MEMORY;
		goto exit_set_dev_entry;
	}

	if (ioctl->RequestLen < sizeof(*dev_entry) || 
	    dma_buf->buf_len < sizeof(*fw_ddb_entry))
	{
		printk("qla4extioctl_set_device_entry_iscsi: memory area too small\n");
		ioctl->Status = EXT_STATUS_BUFFER_TOO_SMALL;
		goto exit_set_dev_entry;
	}

	/*
	 * Copy the IOCTL EXT_DEVICE_ENTRY_ISCSI buffer from the user's data space
	 */
	if (copy_from_user((void *)dev_entry, (void *)LSL(ioctl->RequestAdr), ioctl->RequestLen) != 0)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_set_device_entry_iscsi: unable to copy data from user's memory area\n"));
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_set_dev_entry;
	}

	/*
	 * Transfer data from IOCTL's EXT_DEVICE_ENTRY_ISCSI buffer to Fw's DEVDB_ENTRY buffer
	 */
	fw_ddb_entry = dma_buf->virt_addr;
	memset(fw_ddb_entry, 0, sizeof(*fw_ddb_entry));

	fw_ddb_entry->options          = dev_entry->Options;
	fw_ddb_entry->control          = dev_entry->Control;
	fw_ddb_entry->TSID             = dev_entry->TargetSessID;
	fw_ddb_entry->exeCount         = dev_entry->ExeCount;
	fw_ddb_entry->ddbLink          = dev_entry->DDBLink;
	memcpy(fw_ddb_entry->ISID,       dev_entry->InitiatorSessID, sizeof(dev_entry->InitiatorSessID));
	memcpy(fw_ddb_entry->userID,     dev_entry->UserID,          sizeof(dev_entry->UserID));
	memcpy(fw_ddb_entry->password,   dev_entry->Password,        sizeof(dev_entry->Password));

	fw_ddb_entry->exeThrottle      = dev_entry->DeviceInfo.ExeThrottle;
	fw_ddb_entry->initMarkerlessInt= dev_entry->DeviceInfo.InitMarkerlessInt;
	fw_ddb_entry->retryCount       = dev_entry->DeviceInfo.RetryCount;
	fw_ddb_entry->retryDelay       = dev_entry->DeviceInfo.RetryDelay;
	fw_ddb_entry->iSCSIOptions     = dev_entry->DeviceInfo.iSCSIOptions;
	fw_ddb_entry->TCPOptions       = dev_entry->DeviceInfo.TCPOptions;
	fw_ddb_entry->IPOptions        = dev_entry->DeviceInfo.IPOptions;
	fw_ddb_entry->maxPDUSize       = dev_entry->DeviceInfo.MaxPDUSize;
	fw_ddb_entry->firstBurstSize   = dev_entry->DeviceInfo.FirstBurstSize;
	fw_ddb_entry->logoutMinTime    = dev_entry->DeviceInfo.LogoutMinTime;
	fw_ddb_entry->logoutMaxTime    = dev_entry->DeviceInfo.LogoutMaxTime;
	fw_ddb_entry->maxOutstndngR2T  = dev_entry->DeviceInfo.MaxOutstandingR2T;
	fw_ddb_entry->keepAliveTimeout = dev_entry->DeviceInfo.KeepAliveTimeout;
	fw_ddb_entry->portNumber       = dev_entry->DeviceInfo.PortNumber;
	fw_ddb_entry->maxBurstSize     = dev_entry->DeviceInfo.MaxBurstSize;
	fw_ddb_entry->taskMngmntTimeout= dev_entry->DeviceInfo.TaskMgmtTimeout;
	memcpy(fw_ddb_entry->targetAddr, dev_entry->DeviceInfo.TargetAddr, sizeof(dev_entry->DeviceInfo.TargetAddr));

	memcpy(fw_ddb_entry->ipAddr,     dev_entry->EntryInfo.IPAddr.IPAddress, sizeof(dev_entry->EntryInfo.IPAddr.IPAddress));
	memcpy(fw_ddb_entry->iscsiName,  dev_entry->EntryInfo.iSCSIName,        sizeof(dev_entry->EntryInfo.iSCSIName));
	memcpy(fw_ddb_entry->iSCSIAlias, dev_entry->EntryInfo.Alias,            sizeof(dev_entry->EntryInfo.Alias));

	/* 
	 * Make the IOCTL call
	 */
	if (qla4000_set_ddb_entry(ha, ioctl->Instance, fw_ddb_entry) != QLA_SUCCESS)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_set_device_entry_iscsi: SET DDB Entry failed\n"));
		ioctl->Status = EXT_STATUS_ERR;
		goto exit_set_dev_entry;
	}

	ioctl->Status = EXT_STATUS_OK;
	status = QLA_SUCCESS;

exit_set_dev_entry:
	if (dev_entry) kfree(dev_entry);
	LEAVE("qla4extioctl_set_device_entry_iscsi");
	return status;
}

static int qla4extioctl_set_init_fw_iscsi(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int               status   = QLA_ERROR;
	dma_buf_t         *dma_buf = &ha->dma_buf;
	EXT_INIT_FW_ISCSI *init_fw = kmalloc(sizeof(*init_fw), GFP_ATOMIC);
	INIT_FW_CTRL_BLK  *init_fw_cb;
	static uint32_t mbox_cmd[MBOX_REG_COUNT];
	static uint32_t mbox_sts[MBOX_REG_COUNT];

	ENTER("qla4extioctl_set_init_fw_iscsi");
	spin_lock(&dma_buf->spinlock);

	if (!dma_buf->virt_addr || !dma_buf->phys_addr || !init_fw || 
	    (dma_buf->buf_len < sizeof(INIT_FW_CTRL_BLK)) ||
	    (ioctl->RequestLen < sizeof(EXT_INIT_FW_ISCSI))
	   )
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_set_init_fw_iscsi: requst buffer too small\n"));
		ioctl->Status = EXT_STATUS_BUFFER_TOO_SMALL;
		goto exit_set_init_fw;
	}

	/*
	 * Copy the data from the user's buffer
	 */
	if (copy_from_user(init_fw, (void*)LSL(ioctl->RequestAdr), sizeof(*init_fw)) != 0)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_set_init_fw_iscsi: unable to copy data to user's memory area\n"));
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_set_init_fw;
	}

	/* 
	 * First get Initialize Firmware Control Block, so as not to destroy unaffected data
	 */
	memset(&mbox_cmd, 0, sizeof(mbox_cmd));
	mbox_cmd[0] = MBOX_CMD_GET_INIT_FW_CTRL_BLOCK;
	mbox_cmd[2] = LSL(dma_buf->phys_addr);
	mbox_cmd[3] = MSL(dma_buf->phys_addr);

	if (qla4000_mailbox_command(ha, 4, 1, &mbox_cmd[0], &mbox_sts[0]) == QLA_ERROR)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_get_init_fw_iscsi: command failed \n"));
		ioctl->Status = EXT_STATUS_ERR;
		ioctl->DetailStatus = mbox_sts[0];
		goto exit_set_init_fw;
	}

	/*
	 * Transfer Data from Local buffer to DMA buffer
	 */
	init_fw_cb = (INIT_FW_CTRL_BLK *)dma_buf->virt_addr;

	init_fw_cb->Version         = init_fw->Version;
	init_fw_cb->FwOptions       = init_fw->FWOptions;
	init_fw_cb->AddFwOptions    = init_fw->AddFWOptions;
	init_fw_cb->WakeupThreshold = init_fw->WakeupThreshold;
	memcpy(init_fw_cb->IPAddr,          init_fw->IPAddr.IPAddress,     EXT_DEF_IP_ADDR_SIZE);
	memcpy(init_fw_cb->SubnetMask,      init_fw->SubnetMask.IPAddress, EXT_DEF_IP_ADDR_SIZE);
	memcpy(init_fw_cb->GatewayIPAddr,   init_fw->Gateway.IPAddress,    EXT_DEF_IP_ADDR_SIZE);
	memcpy(init_fw_cb->DNSIPAddr, init_fw->DNSConfig.IPAddr.IPAddress, EXT_DEF_IP_ADDR_SIZE);
	memcpy(init_fw_cb->Alias,           init_fw->Alias,                EXT_DEF_ISCSI_ALIAS_LEN);
	memcpy(init_fw_cb->iSCSINameString, init_fw->iSCSIName,            EXT_DEF_ISCSI_NAME_LEN);

	init_fw_cb->ExecThrottle       = init_fw->DeviceInfo.ExeThrottle;
	init_fw_cb->InitMarkerlessInt  = init_fw->DeviceInfo.InitMarkerlessInt;
	init_fw_cb->RetryCount         = init_fw->DeviceInfo.RetryCount;
	init_fw_cb->RetryDelay         = init_fw->DeviceInfo.RetryDelay;
	init_fw_cb->iSCSIOptions       = init_fw->DeviceInfo.iSCSIOptions;
	init_fw_cb->TCPOptions         = init_fw->DeviceInfo.TCPOptions;
	init_fw_cb->IPOptions          = init_fw->DeviceInfo.IPOptions;
	init_fw_cb->MaxPDUSize         = init_fw->DeviceInfo.MaxPDUSize;
	init_fw_cb->FirstBurstSize     = init_fw->DeviceInfo.FirstBurstSize;
	init_fw_cb->LogOutMinTime      = init_fw->DeviceInfo.LogoutMinTime;
	init_fw_cb->LogOutMaxTime      = init_fw->DeviceInfo.LogoutMaxTime;
	init_fw_cb->LogOutMaxTime      = init_fw->DeviceInfo.LogoutMaxTime;
	init_fw_cb->MaxOutStndngR2T    = init_fw->DeviceInfo.MaxOutstandingR2T;
	init_fw_cb->KeepAliveTimeout   = init_fw->DeviceInfo.KeepAliveTimeout;
	init_fw_cb->PortNumber         = init_fw->DeviceInfo.PortNumber;
	init_fw_cb->MaxBurstSize       = init_fw->DeviceInfo.MaxBurstSize;
	//init_fw_cb->?                = init_fw->DeviceInfo.TaskMgmtTimeout;
	memcpy(init_fw_cb->TargAddr, init_fw->DeviceInfo.TargetAddr, EXT_DEF_ISCSI_TADDR_SIZE); 

	/*
	 * Send mailbox command
	 */

	memset(&mbox_cmd, 0, sizeof(mbox_cmd));
	mbox_cmd[0] = MBOX_CMD_INITIALIZE_FIRMWARE;
	mbox_cmd[2] = LSL(dma_buf->phys_addr);
	mbox_cmd[3] = MSL(dma_buf->phys_addr);

	if (qla4000_mailbox_command(ha, 4, 1, &mbox_cmd[0], &mbox_sts[0]) == QLA_ERROR)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_set_init_fw_iscsi: command failed \n"));
		ioctl->Status = EXT_STATUS_ERR;
		ioctl->DetailStatus = mbox_sts[0];
		goto exit_set_init_fw;
	}

	ioctl->Status = EXT_STATUS_OK;
	status = QLA_SUCCESS;

	QL4PRINT(QLP10, printk("EXT_INIT_FW_ISCSI structure:\n"));
	qla4000_dump_bytes(QLP10, init_fw, sizeof(*init_fw));

exit_set_init_fw:
	if (init_fw) kfree(init_fw);
	spin_unlock(&dma_buf->spinlock);
	LEAVE("qla4extioctl_set_init_fw_iscsi");
	return status;
}

static int qla4extioctl_set_data(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	switch (ioctl->SubCode)
	{
	case EXT_SC_RST_STATISTICS_GEN:
		return qla4extioctl_rst_statistics_gen(ha, ioctl);

	case EXT_SC_RST_STATISTICS_ISCSI:
		return qla4extioctl_rst_statistics_iscsi(ha, ioctl);

	case EXT_SC_SET_DEVICE_ENTRY_ISCSI:
		return qla4extioctl_set_device_entry_iscsi(ha, ioctl);

	case EXT_SC_SET_INIT_FW_ISCSI:
		return qla4extioctl_set_init_fw_iscsi(ha, ioctl);

	default:
		QL4PRINT(QLP1|QLP4, printk("qla4000_ioctl: unsupported set data sub-command code (%X)\n", ioctl->SubCode));
		ioctl->Status = EXT_STATUS_UNSUPPORTED_SUBCODE;
		return(-EINVAL);
	}

}

/*
 * qla4000_ioctl_sleep_done
 *
 * Description:
 *   This is the callback function to wakeup ioctl completion semaphore
 *   for the ioctl request that is waiting.
 *
 * Input:
 *   sem - pointer to the ioctl completion semaphore.
 *
 * Returns:
 */
STATIC void
qla4000_ioctl_sleep_done (struct semaphore * sem)
{
	ENTER("qla4000_ioctl_sleep_done");

	if (sem != NULL){
		QL4PRINT(QLP6, printk("qla4000_ioctl_sleep_done: wake up sem.\n"));
		up(sem);
	}

	LEAVE("qla4000_ioctl_sleep_done");
}

/*
 * qla4000_ioctl_sem_init
 *
 * Description:
 *   Initialize the ioctl timer and semaphore used to wait for passthru
 *   completion.
 *
 * Input:
 *   ha - pointer to adapter_state_t structure used for initialization.
 *
 * Returns:
 *   None.
 */
void
qla4000_ioctl_sem_init (scsi_qla_host_t *ha)
{
	init_timer(&(ha->ioctl_cmpl_timer));
	ha->ioctl_cmpl_timer.data = (unsigned long)&ha->ioctl_cmpl_sem;
	ha->ioctl_cmpl_timer.function =
	    (void (*)(unsigned long))qla4000_ioctl_sleep_done;
}

/*
 * qla4extioctl_scsi_pass_done
 *
 * Description:
 *   Resets ioctl progress flag and wakes up the ioctl completion semaphore.
 *
 * Input:
 *   cmd - pointer to the passthru Scsi cmd structure which has completed.
 *
 * Returns:
 */
void qla4000_scsi_pass_done(Scsi_Cmnd *cmd)
{
	scsi_qla_host_t *ha = (scsi_qla_host_t *) cmd->host->hostdata;

	ENTER("qla4000_scsi_pass_done");

	ha->ioctl_scsi_pass_in_progress = FALSE;
	ha->ioctl_tov = 0;

	up(&ha->ioctl_cmpl_sem);

	LEAVE("qla4000_scsi_pass_done");
	return;
}

/*
 * qla4extioctl_scsi_passthru
 *
 * 
 * dma_buf                              
 * ioctl->RequestBuf 
 * ioctl->ResponseBuf 
 *    +-------------------------+
 *    | EXT_SCSI_PASSTHRU_ISCSI |
 *    +-------------------------+
 *    | [SCSI READ|WRITE data]  |
 *    +-------------------------+
 */
static int qla4extioctl_scsi_passthru(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int 			status = QLA_ERROR;
	ddb_entry_t		*ddb_entry;
	int 			i;
	static EXT_SCSI_PASSTHRU_ISCSI scsi_pass;
	static Scsi_Device	scsi_device;
	static Scsi_Cmnd   	cmd;
	unsigned long flags;

	ENTER("qla4extioctl_scsi_passthru");
	spin_lock(&ha->dma_buf.spinlock);

	memset(&scsi_device, 0, sizeof(scsi_device));
	memset(&scsi_pass, 0, sizeof(scsi_pass));
	memset(&cmd, 0, sizeof(cmd));

	/* ---- Allocate larger DMA buffer, if neccessary ---- */
	if (ha->dma_buf.buf_len < ioctl->ResponseLen &&
        qla4000_resize_dma_buf(ha, &ha->dma_buf, ioctl->ResponseLen) != QLA_SUCCESS)
	{
		QL4PRINT(
			QLP2|QLP4, 
			printk("qla4extioctl_scsi_passthru: ERROR cannot allocate requested DMA buffer size 0x%x.\n", 
			ioctl->ResponseLen));
		ioctl->Status = EXT_STATUS_NO_MEMORY;
		goto error_exit_scsi_pass;
	}

	memset(ha->dma_buf.virt_addr, 0, ha->dma_buf.buf_len);

	/* ---- Get passthru structure from user space ---- */
	if (copy_from_user(&scsi_pass, (void *)LSL(ioctl->RequestAdr), sizeof(scsi_pass)) != 0)
	{
		QL4PRINT(QLP2|QLP4, printk("qla4extioctl_scsi_passthru: unable to copy passthru struct from user's memory area.\n"));
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto error_exit_scsi_pass;
	}

	QL4PRINT(QLP10, printk("qla4extioctl_scsi_passthru: incoming  EXT_SCSI_PASSTHRU_ISCSI structure: \n"));
	qla4000_dump_bytes(QLP10, (void *)LSL(ioctl->RequestAdr), sizeof(EXT_SCSI_PASSTHRU_ISCSI));

	/* ---- Make sure device exists ---- */
	ddb_entry = qla4000_lookup_ddb_by_SCSIID(ha, scsi_pass.Addr.Bus, scsi_pass.Addr.Target);
	if (ddb_entry == NULL)
	{
		QL4PRINT(QLP2|QLP4, printk("qla4extioctl_scsi_passthru: invalid device (b%d,t%d,l%d) specified.\n",
				    scsi_pass.Addr.Bus, scsi_pass.Addr.Target, scsi_pass.Addr.Lun));
		ioctl->Status = EXT_STATUS_DEV_NOT_FOUND;
		goto error_exit_scsi_pass;
	}

	/* ---- Make sure device is in an active state ---- */
	if (ddb_entry->device_state != DDB_DS_SESSION_ACTIVE)
	{
		QL4PRINT(QLP2|QLP4, printk("qla4extioctl_scsi_passthru: device (b%d,t%d) not in active state.\n",
				    scsi_pass.Addr.Bus, scsi_pass.Addr.Target));
		ioctl->Status = EXT_STATUS_DEVICE_NOT_READY;
		goto error_exit_scsi_pass;
	}

	/* ---- Fill in the SCSI command structure ---- */
	cmd.channel         = scsi_pass.Addr.Bus;
	cmd.target          = scsi_pass.Addr.Target;
	cmd.lun             = scsi_pass.Addr.Lun;
	cmd.request_buffer  = ha->dma_buf.virt_addr;
	cmd.request_bufflen = ioctl->RequestLen;
	cmd.device  	    = &scsi_device;
    cmd.host       	    = ha->host;
	cmd.scsi_done       = qla4000_scsi_pass_done;
	CMD_TIMEOUT(&cmd)	= IOCTL_PASSTHRU_TOV * HZ;

	if (scsi_pass.CdbLength == 6 || 
	    scsi_pass.CdbLength == 10 ||
	    scsi_pass.CdbLength == 12)
	{
		cmd.cmd_len = scsi_pass.CdbLength;
	}
	else 
	{
		QL4PRINT(QLP2|QLP4, printk("qla4extioctl_scsi_passthru: Unsupported CDB length 0x%x \n", cmd.cmd_len));
		ioctl->Status = EXT_STATUS_INVALID_PARAM;
		goto error_exit_scsi_pass;
	}

	if (scsi_pass.Direction == EXT_DEF_SCSI_PASSTHRU_DATA_IN)
		cmd.sc_data_direction = SCSI_DATA_READ;
	else if (scsi_pass.Direction ==  EXT_DEF_SCSI_PASSTHRU_DATA_OUT)
	{
		/* sending user data from ioctl->ResponseAddr to SCSI command buffer */
		if (copy_from_user(cmd.request_buffer, (void*)LSL(ioctl->ResponseAdr), cmd.request_bufflen) != 0)
		{
			QL4PRINT(QLP2|QLP4, printk("qla4extioctl_scsi_passthru: unable to copy write buffer from user's memory area.\n"));
			ioctl->Status = EXT_STATUS_COPY_ERR;
			goto error_exit_scsi_pass;
		}
		cmd.sc_data_direction = SCSI_DATA_WRITE;
	}
	else
	{
		cmd.sc_data_direction = SCSI_DATA_NONE;
		cmd.request_buffer  = 0;
		cmd.request_bufflen = 0;
	}


	memcpy(cmd.cmnd,      scsi_pass.Cdb, cmd.cmd_len);
	memcpy(cmd.data_cmnd, scsi_pass.Cdb, cmd.cmd_len);

	QL4PRINT(QLP4, printk("qla4extioctl_scsi_passthru (%d:%d:%d:%d): CDB = ",
			    ha->host_no, cmd.channel, cmd.target, cmd.lun));
	for (i = 0; i < cmd.cmd_len; i++)
            QL4PRINT(QLP4, printk("%02X ", cmd.cmnd[i]));
	QL4PRINT(QLP4, printk("\n"));


	/* ---- prepare for receiving completion ---- */
	ha->ioctl_scsi_pass_in_progress = TRUE;
	ha->ioctl_tov = CMD_TIMEOUT(&cmd);

	qla4000_ioctl_sem_init(ha);
	CMD_COMPL_STATUS(&cmd)  = (int) IOCTL_INVALID_STATUS;
	CMD_PASSTHRU_TYPE(&cmd) = TRUE;

	/* ---- send command to adapter ---- */
	QL4PRINT(QLP4, printk("qla4extioctl_scsi_passthru (%d:%d:%d:%d): sending command.\n",
			    ha->host_no, cmd.channel, cmd.target, cmd.lun));
	spin_lock_irqsave(&io_request_lock, flags);
	qla4000_queuecommand(&cmd, (void *) qla4000_scsi_pass_done);

	ha->ioctl_cmpl_timer.expires = jiffies + ha->ioctl_tov;
	add_timer(&ha->ioctl_cmpl_timer);

	spin_unlock_irqrestore(&io_request_lock, flags);
	down(&ha->ioctl_cmpl_sem);

	/*******************************************************
	 *						       *
	 *             Passthru Completion                     *
	 *						       *
	 *******************************************************/
	del_timer(&ha->ioctl_cmpl_timer);

	/* ---- check for timeout --- */
	if (ha->ioctl_scsi_pass_in_progress == 1)
	{
		QL4PRINT(QLP2|QLP4, printk("qla4extioctl_scsi_passthru: ERROR = command timeout.\n"));
		ioctl->Status = EXT_STATUS_ERR;
		goto error_exit_scsi_pass;
	}

	/* ---- check for command completion --- */
	if (CMD_COMPL_STATUS(&cmd) == IOCTL_INVALID_STATUS)
	{
		QL4PRINT(QLP2|QLP6, printk("qla4extioctl_scsi_passthru: ERROR = command not completed.\n"));
		ioctl->Status = EXT_STATUS_ERR;
		goto error_exit_scsi_pass;
	}
	else if (CMD_HOST_STATUS(&cmd) == DID_OK)
	{
		ioctl->Status = EXT_STATUS_OK;
	}
	else if (CMD_COMPL_STATUS(&cmd) == SCS_DATA_UNDERRUN)
	{
		QL4PRINT(QLP2|QLP6, printk("qla4extioctl_scsi_passthru: Data underrun.  Resid = 0x%x\n",
			    CMD_RESID_LEN(&cmd)));
		ioctl->Status = EXT_STATUS_DATA_UNDERRUN;
		ioctl->DetailStatus = CMD_RESID_LEN(&cmd);
	}
	else if (CMD_COMPL_STATUS(&cmd) == SCS_DATA_OVERRUN)
	{
		QL4PRINT(QLP2|QLP6, printk("qla4extioctl_scsi_passthru: Data overrun.  Resid = 0x%x\n",
			    CMD_RESID_LEN(&cmd)));
		ioctl->Status = EXT_STATUS_DATA_OVERRUN;
		ioctl->DetailStatus = CMD_RESID_LEN(&cmd);
	}
	else
	{
		QL4PRINT(QLP2|QLP6, printk("qla4extioctl_scsi_passthru: Command completed in ERROR. cs=%04x, ss=%-4x\n",
			    CMD_COMPL_STATUS(&cmd), CMD_SCSI_STATUS(&cmd)));

		if (CMD_SCSI_STATUS(&cmd) != SCSI_GOOD)
			ioctl->Status = EXT_STATUS_SCSI_STATUS;
		else 
			ioctl->Status = EXT_STATUS_ERR;
	}

	/* --- Return info from status entry --- */
	ioctl->VendorSpecificStatus[VS_STATUS_SCSI_STATUS_INDEX]      = CMD_SCSI_STATUS(&cmd);
	ioctl->VendorSpecificStatus[VS_STATUS_SCSI_COMPLETION_INDEX]  = CMD_COMPL_STATUS(&cmd);
	ioctl->VendorSpecificStatus[VS_STATUS_SENSE_BYTE_COUNT_INDEX] = CMD_ACTUAL_SNSLEN(&cmd);
	ioctl->VendorSpecificStatus[VS_STATUS_HOST_STATUS_INDEX]      = CMD_HOST_STATUS(&cmd);
	ioctl->VendorSpecificStatus[VS_STATUS_ISCSI_RESPONSE_INDEX]   = CMD_ISCSI_RESPONSE(&cmd);
	ioctl->VendorSpecificStatus[VS_STATUS_STATE_FLAGS_INDEX]      = CMD_STATE_FLAGS(&cmd);

	scsi_pass.Reserved[0] = (uint8_t) CMD_SCSI_STATUS(&cmd);  
	scsi_pass.Reserved[1] = (uint8_t) CMD_COMPL_STATUS(&cmd); 
	scsi_pass.Reserved[2] = (uint8_t) CMD_ACTUAL_SNSLEN(&cmd);
	scsi_pass.Reserved[3] = (uint8_t) CMD_HOST_STATUS(&cmd);  

	/* ---- Copy SCSI Passthru structure with updated sense buffer to user space ---- */
	if (CMD_ACTUAL_SNSLEN(&cmd))
	{
		memcpy(scsi_pass.SenseData, CMD_SNSP(&cmd), 
		       MIN(CMD_ACTUAL_SNSLEN(&cmd), sizeof(scsi_pass.SenseData)));
	}

	if (copy_to_user((void *)LSL(ioctl->ResponseAdr),
                     &scsi_pass, 
			 sizeof(EXT_SCSI_PASSTHRU_ISCSI)) != 0)
	{
		QL4PRINT(QLP2|QLP4, printk("qla4extioctl_scsi_passthru: unable to copy passthru struct to user's memory area.\n"));
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto error_exit_scsi_pass;
	}

	QL4PRINT(QLP10, printk("qla4extioctl_scsi_passthru: outgoing EXT_SCSI_PASSTHRU_ISCSI structure:\n"));
	qla4000_dump_bytes(QLP10, (void *)LSL(ioctl->ResponseAdr), sizeof(EXT_SCSI_PASSTHRU_ISCSI));

   /* ---- Copy SCSI READ data from SCSI command buffer to user space ---- */
	if (scsi_pass.Direction == EXT_DEF_SCSI_PASSTHRU_DATA_IN)
	{
		void    *xfer_ptr = (void *)LSL(ioctl->ResponseAdr) + sizeof(EXT_SCSI_PASSTHRU_ISCSI);
		uint32_t xfer_len = ioctl->ResponseLen - sizeof(EXT_SCSI_PASSTHRU_ISCSI);


		/* Update ResponseLen if a data underrun occurred */
		if (CMD_COMPL_STATUS(&cmd) == SCS_DATA_UNDERRUN && CMD_RESID_LEN(&cmd))
			xfer_len -= CMD_RESID_LEN(&cmd);

		
		if (copy_to_user(xfer_ptr, cmd.request_buffer, xfer_len) != 0)
		{
			QL4PRINT(QLP2|QLP4, printk("qla4extioctl_scsi_passthru: unable to copy READ data to user's memory area.\n"));
			ioctl->Status = EXT_STATUS_COPY_ERR;
			goto error_exit_scsi_pass;
		}

		QL4PRINT(QLP10, printk("qla4extioctl_scsi_passthru: outgoing READ data:  (0x%p)\n", xfer_ptr));
		qla4000_dump_bytes(QLP10, xfer_ptr, xfer_len);
	}

	goto exit_scsi_pass;

error_exit_scsi_pass:
	ioctl->ResponseLen = 0;

exit_scsi_pass:
	spin_unlock(&ha->dma_buf.spinlock);
	status = ioctl->Status;
	LEAVE("qla4extioctl_scsi_passthru");
	return status;
}

static int qla4extioctl_iscsi_passthru(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = QLA_ERROR;
	ioctl->Status = EXT_STATUS_UNSUPPORTED_SUBCODE;
	ENTER("qla4extioctl_iscsi_passthru");
	LEAVE("qla4extioctl_iscsi_passthru");
	return status;
}

static int qla4extioctl_get_hbacnt(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = QLA_ERROR;
	uint32_t hba_count;

	ENTER("qla4extioctl_get_hbacnt");

	hba_count = qla4000_get_hba_count();
	if (put_user(hba_count, (uint16_t *) LSL(ioctl->ResponseAdr)) != 0)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4extioctl_get_hbacnt: failed to copy data\n"));
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_get_hbacnt;
	}

	QL4PRINT(QLP10, printk("qla4extioctl_get_hbacnt: hbacnt is %d\n", hba_count));
	status = QLA_SUCCESS;
	ioctl->Status = EXT_STATUS_OK;

exit_get_hbacnt:

	LEAVE("qla4extioctl_get_hbacnt");
	return status;
}

static int qla4intioctl_logout_iscsi(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = QLA_ERROR;
	INT_LOGOUT_ISCSI logout;
	
	ENTER("qla4intioctl_logout_iscsi");

	if (ioctl->RequestLen > sizeof(INT_LOGOUT_ISCSI))
	{
			QL4PRINT(QLP1|QLP4, printk("qla4intioctl_logout_iscsi: memory area too small\n"));
			ioctl->Status = EXT_STATUS_BUFFER_TOO_SMALL;
			goto exit_logout;
	}

	/* --- Copy logout structure from user space --- */
	if (copy_from_user(&logout, (void*)LSL(ioctl->RequestAdr), sizeof(logout)) != 0)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4intioctl_logout_iscsi: unable to copy data from user's memory area\n"));
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_logout;
	}

	/* --- Execute command --- */
	if (logout.Options == INT_DEF_CLOSE_SESSION)
	{
			if ((status = qla4000_logout_device(ha, logout.TargetID, logout.ConnectionID)) == QLA_SUCCESS)
			{
					QL4PRINT(QLP6, printk("qla4intioctl_logout_iscsi: CLOSE_SESSION SUCCEEDED!\n"));
					ioctl->Status = EXT_STATUS_OK;
			}
			else 
			{
					QL4PRINT(QLP2|QLP6, printk("qla4intioctl_logout_iscsi: CLOSE_SESSION FAILED!\n"));
					ioctl->Status = EXT_STATUS_ERR;
			}
	}
	else if (logout.Options == INT_DEF_RELOGIN_CONNECTION)
	{
			if ((status = qla4000_login_device(ha, logout.TargetID, logout.ConnectionID)) == QLA_SUCCESS)
			{
					QL4PRINT(QLP6, printk("qla4intioctl_logout_iscsi: RELOGIN_CONNECTION SUCCEEDED!\n"));
					ioctl->Status = EXT_STATUS_OK;
			}
			else 
			{
					QL4PRINT(QLP2|QLP6, printk("qla4intioctl_logout_iscsi: RELOGIN_CONNECTION FAILED!\n"));
					ioctl->Status = EXT_STATUS_ERR;
			}
	}

exit_logout:
	LEAVE("qla4intioctl_logout_iscsi");
	return status;
}

static int qla4intioctl_ping(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = QLA_ERROR;
	ioctl->Status = EXT_STATUS_UNSUPPORTED_SUBCODE;
	ENTER("qla4intioctl_ping");
	LEAVE("qla4intioctl_ping");
	return status;
}

static int qla4intioctl_get_flash(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int         	 status = QLA_ERROR;
	dma_buf_t        dma_buf;
	INT_ACCESS_FLASH *access_flash;
	static uint32_t         mbox_cmd[MBOX_REG_COUNT];
	static uint32_t         mbox_sts[MBOX_REG_COUNT];

	ENTER("qla4intioctl_get_flash");

	/*
	 * Allocate DMA memory
	 */
	if (qla4000_alloc_physical_memory(ha->pdev, &dma_buf, sizeof(*access_flash)) == QLA_ERROR)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4intioctl_get_flash: unable to allocate dma memory\n"));
		ioctl->Status = EXT_STATUS_NO_MEMORY;
		goto exit_get_flash;
	}

	/*
	 * Copy user's data to local DMA buffer
	 */
	access_flash = dma_buf.virt_addr;
	if (copy_from_user(access_flash, (void*)LSL(ioctl->RequestAdr), sizeof(*access_flash)) != 0)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4intioctl_get_flash: unable to copy data from user's memory area\n"));
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_get_flash;
	}

	/*
	 * Issue Mailbox Command
	 */
	memset(&mbox_cmd, 0, sizeof(mbox_cmd));
	mbox_cmd[0] = MBOX_CMD_READ_FLASH;
	mbox_cmd[1] = LSL(dma_buf.phys_addr + OFFSET(access_flash->FlashData));
	mbox_cmd[2] = MSL(dma_buf.phys_addr + OFFSET(access_flash->FlashData));
	mbox_cmd[3] = access_flash->DataOffset;
	mbox_cmd[4] = access_flash->DataLen;

	if (qla4000_mailbox_command(ha, 5, 1, &mbox_cmd[0], &mbox_sts[0]) == QLA_ERROR)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4intioctl_get_flash: command failed \n"));
		ioctl->Status = EXT_STATUS_ERR;
		ioctl->DetailStatus = mbox_sts[0];
		goto exit_get_flash;
	}

	/*
	 * Copy local DMA buffer to user's response data area
	 */
	if (copy_to_user((void*)LSL(ioctl->ResponseAdr), dma_buf.virt_addr, dma_buf.buf_len) != 0)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4intioctl_get_flash: unable to copy data to user's memory area\n"));
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_get_flash;
	}

	ioctl->Status = EXT_STATUS_OK;
	ioctl->ResponseLen = access_flash->DataLen;
	QL4PRINT(QLP10, printk("INT_ACCESS_FLASH buffer (1st 60h bytes only):\n")); 
	qla4000_dump_bytes(QLP10, dma_buf.virt_addr, 0x60);
	status = QLA_SUCCESS;

exit_get_flash:
	/*
	 * Free DMA Memory
	 */

	if (dma_buf.virt_addr)
		qla4000_free_physical_memory(ha->pdev, &dma_buf);

	LEAVE("qla4intioctl_get_flash");
	return status;
}

static int qla4intioctl_get_driver_debug_level(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = QLA_ERROR;
	uint32_t dbg_level;
	
	ENTER("qla4intioctl_get_driver_debug_level");

	if (qla4000_get_debug_level(&dbg_level) == QLA_ERROR)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4intioctl_get_driver_debug_level: unable to set debug level, debug driver not loaded!\n"));
		ioctl->Status = EXT_STATUS_ERR;
		goto exit_get_driver_debug_level;
	}

	if (put_user(dbg_level, (uint32_t *) LSL(ioctl->ResponseAdr)) != 0)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4intioctl_set_driver_debug_level: failed to copy data\n"));
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_get_driver_debug_level;
	}

	QL4PRINT(QLP3|QLP4, printk("qla4intioctl_get_driver_debug_level: debug level is %04x\n", dbg_level));
	status = QLA_SUCCESS;
	ioctl->Status = EXT_STATUS_OK;

exit_get_driver_debug_level:

	LEAVE("qla4intioctl_get_driver_debug_level");
	return status;
}

static int qla4intioctl_get_data(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	switch (ioctl->SubCode)
	{
	case INT_SC_GET_FLASH:
		return qla4intioctl_get_flash(ha, ioctl);

	case INT_SC_GET_DRIVER_DEBUG_LEVEL:
		return qla4intioctl_get_driver_debug_level(ha, ioctl);

	default:
		QL4PRINT(QLP1|QLP4, printk("qla4000_ioctl: unsupported internal get data sub-command code (%X)\n", ioctl->SubCode));
		ioctl->Status = EXT_STATUS_UNSUPPORTED_SUBCODE;
		return(-EINVAL);
	}
}

static int qla4intioctl_set_flash(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = QLA_ERROR;
	dma_buf_t        dma_buf;
	INT_ACCESS_FLASH *access_flash;
	static uint32_t         mbox_cmd[MBOX_REG_COUNT];
	static uint32_t         mbox_sts[MBOX_REG_COUNT];

	ENTER("qla4intioctl_set_flash");

	/*
	 * Allocate DMA memory
	 */
	if (qla4000_alloc_physical_memory(ha->pdev, &dma_buf, sizeof(*access_flash)) == QLA_ERROR)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4intioctl_set_flash: unable to allocate dma memory\n"));
		ioctl->Status = EXT_STATUS_NO_MEMORY;
		goto exit_set_flash;
	}

	/*
	 * Copy user's data to local DMA buffer
	 */
	access_flash = dma_buf.virt_addr;
	if (copy_from_user(access_flash, (void*)LSL(ioctl->RequestAdr), sizeof(*access_flash)) != 0)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4intioctl_set_flash: unable to copy data from user's memory area\n"));
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_set_flash;
	}

	/*
	 * Issue Mailbox Command
	 */
	memset(&mbox_cmd, 0, sizeof(mbox_cmd));
	mbox_cmd[0] = MBOX_CMD_WRITE_FLASH;
	mbox_cmd[1] = LSL(dma_buf.phys_addr + OFFSET(access_flash->FlashData));
	mbox_cmd[2] = MSL(dma_buf.phys_addr + OFFSET(access_flash->FlashData));
	mbox_cmd[3] = access_flash->DataOffset;
	mbox_cmd[4] = access_flash->DataLen;
	mbox_cmd[5] = access_flash->Options;

	if (qla4000_mailbox_command(ha, 6, 2, &mbox_cmd[0], &mbox_sts[0]) == QLA_ERROR)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4intioctl_set_flash: command failed \n"));
		ioctl->Status = EXT_STATUS_ERR;
		ioctl->DetailStatus = mbox_sts[0];
		ioctl->VendorSpecificStatus[0] = mbox_sts[1];
		goto exit_set_flash;
	}

	ioctl->Status = EXT_STATUS_OK;
	ioctl->ResponseLen = access_flash->DataLen;
	QL4PRINT(QLP10, printk("INT_ACCESS_FLASH buffer (1st 60h bytes only):\n")); 
	qla4000_dump_bytes(QLP10, dma_buf.virt_addr, 0x60);
	status = QLA_SUCCESS;

exit_set_flash:
	/*
	 * Free DMA Memory
	 */
	if (dma_buf.virt_addr)
		qla4000_free_physical_memory(ha->pdev, &dma_buf);

	LEAVE("qla4intioctl_set_flash");
	return status;
}

static int qla4intioctl_set_driver_debug_level(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = QLA_ERROR;
	uint32_t dbg_level;

	ENTER("qla4intioctl_set_driver_debug_level");

	if (get_user(dbg_level, (uint32_t *) LSL(ioctl->RequestAdr)) != 0)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4intioctl_set_driver_debug_level: failed to copy data\n"));
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_set_driver_debug_level;
	}

	if (qla4000_set_debug_level(dbg_level) == QLA_ERROR)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4intioctl_set_driver_debug_level: unable to set debug level, debug driver not loaded!\n"));
		ioctl->Status = EXT_STATUS_ERR;
		goto exit_set_driver_debug_level;
	}

	QL4PRINT(QLP3|QLP4, printk("qla4intioctl_set_driver_debug_level: debug level set to 0x%04X\n", dbg_level));
	status = QLA_SUCCESS;
	ioctl->Status = EXT_STATUS_OK;

exit_set_driver_debug_level:

	LEAVE("qla4intioctl_set_driver_debug_level");
	return status;
}

static int qla4intioctl_set_data(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	switch (ioctl->SubCode)
	{
	case INT_SC_SET_FLASH:
		return qla4intioctl_set_flash(ha, ioctl);

	case INT_SC_SET_DRIVER_DEBUG_LEVEL:
		return qla4intioctl_set_driver_debug_level(ha, ioctl);

	default:
		QL4PRINT(QLP1|QLP4, printk("qla4000_ioctl: unsupported internal set data sub-command code (%X)\n", ioctl->SubCode));
		ioctl->Status = EXT_STATUS_UNSUPPORTED_SUBCODE;
		return(-EINVAL);
	}
}

static int qla4intioctl_hba_reset(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = QLA_ERROR;

	ENTER("qla4intioctl_hba_reset");

	switch (ioctl->SubCode)
	{
	case INT_SC_HBA_RESET:
	case INT_SC_FIRMWARE_RESET:
#if 0
		status = qla4000_recover_adapter(ha, 1);
#endif

		set_bit(DPC_RESET_HA_DESTROY_DDB_LIST, &ha->dpc_flags);
		set_bit(AF_SCHEDULE_DPC, &ha->flags);
		do {
			/*Checking to see if reset completed */
			if(test_bit(DPC_RESET_HA_DESTROY_DDB_LIST, 
				    &ha->dpc_flags) == 0)
			{
				status = QLA_SUCCESS;
				break;
			}

			/* wait for 2 seconds */
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(2*HZ);
		} while (1);
		break;

	case INT_SC_TARGET_WARM_RESET:
	case INT_SC_LUN_RESET:
		break;
	}

	if (status == QLA_SUCCESS)
	{
            QL4PRINT(QLP10, printk("qla4intioctl_hba_reset: Succeeded\n"));
		status = QLA_SUCCESS;
		ioctl->Status = EXT_STATUS_OK;
	}
	else 
	{
		QL4PRINT(QLP10, printk("qla4intioctl_hba_reset: FAILED\n"));
		ioctl->Status = EXT_STATUS_ERR;
	}

	LEAVE("qla4intioctl_hba_reset");
	return status;
}

static int qla4intioctl_copy_fw_flash(scsi_qla_host_t *ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = QLA_ERROR;
	INT_COPY_FW_FLASH copy_flash;
	static uint32_t mbox_cmd[MBOX_REG_COUNT];
	static uint32_t mbox_sts[MBOX_REG_COUNT];

	ENTER("qla4intioctl_copy_fw_flash");

	if (copy_from_user(&copy_flash, (void*)LSL(ioctl->RequestAdr), ioctl->RequestLen) != 0)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4intioctl_copy_fw_flash: unable to copy data from user's memory area\n"));
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_copy_flash;

	}

	mbox_cmd[0] = MBOX_CMD_COPY_FLASH;
	mbox_cmd[1] = copy_flash.Options;

	if (qla4000_mailbox_command(ha, 2, 2, &mbox_cmd[0], &mbox_sts[0]) == QLA_SUCCESS)
	{
			QL4PRINT(QLP10, printk("qla4intioctl_copy_fw_flash: Succeeded\n"));
			status = QLA_SUCCESS;
			ioctl->Status = EXT_STATUS_OK;
	}
	else 
	{
			QL4PRINT(QLP10, printk("qla4intioctl_copy_fw_flash: FAILED\n"));
			ioctl->Status = EXT_STATUS_ERR;
			ioctl->DetailStatus = mbox_sts[0];
			ioctl->VendorSpecificStatus[0] = mbox_sts[1];
	}

exit_copy_flash:
	LEAVE("qla4intioctl_copy_fw_flash");
	return status;
}

/*
 * qla4000_ioctl
 *
 * This the main entry point for all ioctl requests
 *
 * Input:
 *    dev - pointer to SCSI device structure
 *	  cmd - internal or external ioctl command code
 *	  arg - pointer to ioctl structure
 *    
 *    Instance field in ioctl structure - to determine which adapter to perform ioctl
 *
 * Output:
 *	 Status field in ioctl structure, if returns QLA_SUCCESS, QLA_ERROR or -EINVAL
 *
 * Returns:
 *	  QLA_SUCCESS 
 *	  QLA_ERROR
 *    -EFAULT      - if the arg pointer is NULL
 *    -EINVAL      - if the command is invalid
 */
int qla4000_ioctl(Scsi_Device *dev, int cmd, void *arg)
{
	EXT_IOCTL_ISCSI *ioctl = &qla4ioctl;
	scsi_qla_host_t *ha;
	int status = QLA_SUCCESS;

	ENTER("qla4000_ioctl");

	/*
	 * Check to see if we can access the ioctl command structure
	 */
	if (!access_ok(VERIFY_WRITE, arg, sizeof(EXT_IOCTL_ISCSI)))
	{
		QL4PRINT(QLP1|QLP4, printk("qla4000_ioctl: NULL EXT_IOCTL_ISCSI buffer\n"));
		status = (-EFAULT);
		goto exit_qla4000_ioctl;
	}

	/*
	 * Copy the ioctl command structure from user space to the local structure
	 */
	copy_from_user((void *) ioctl, arg, sizeof(EXT_IOCTL_ISCSI));
	QL4PRINT(QLP10, printk("EXT_IOCTL_ISCSI structure: \n"));
	qla4000_dump_dwords(QLP10, ioctl, sizeof(*ioctl));

	/*
	 * Get the adapter handle for the corresponding adapter instance
	 */
	ha = qla4000_get_adapter_handle(ioctl->HbaSelect);
	if (ha == NULL)
	{
		QL4PRINT(QLP1|QLP4, printk("qla4000_ioctl: NULL EXT_IOCTL_ISCSI buffer\n"));
		status = QLA_ERROR;
		ioctl->Status = EXT_STATUS_DEV_NOT_FOUND;
		goto exit_qla4000_ioctl;
	}


	down(&ha->ioctl_sem);

	/*
	 * If the adapter is being reset, then we wait until
	 * the recovery is complete
	 */
	if (atomic_read(&ha->adapter_state) == AS_RESET)
	{
		while (1)
		{
			/*Checking to see if reset completed */
			if(atomic_read(&ha->adapter_state) == AS_READY)
				break;

			/* wait for 2 seconds */
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(2*HZ);
		}
	}


	/*
	 * Issue the ioctl command
	 */
	switch (cmd)
	{
	case EXT_CC_QUERY:
		status = qla4extioctl_query(ha, ioctl);
		break;

	case EXT_CC_REG_AEN:
		status = qla4extioctl_reg_aen(ha, ioctl);
		break;

	case EXT_CC_GET_AEN:
		status = qla4extioctl_get_aen(ha, ioctl);
		break;

	case EXT_CC_GET_DATA:
		status = qla4extioctl_get_data(ha, ioctl);
		break;

	case EXT_CC_SET_DATA:
		status = qla4extioctl_set_data(ha, ioctl);
		break;

	case EXT_CC_SEND_SCSI_PASSTHRU:
		status = qla4extioctl_scsi_passthru(ha, ioctl);
		break;

	case EXT_CC_SEND_ISCSI_PASSTHRU:
		status = qla4extioctl_iscsi_passthru(ha, ioctl);
		break;

	case INT_CC_LOGOUT_ISCSI:
		status = qla4intioctl_logout_iscsi(ha, ioctl);
		break;

	case EXT_CC_GET_HBACNT:
		status = qla4extioctl_get_hbacnt(ha, ioctl);
		break;

	case INT_CC_DIAG_PING:
		status = qla4intioctl_ping(ha, ioctl);
		break;

	case INT_CC_GET_DATA:
		status = qla4intioctl_get_data(ha, ioctl);
		break;

	case INT_CC_SET_DATA:
		status = qla4intioctl_set_data(ha, ioctl);
		break;

	case INT_CC_HBA_RESET:
		status = qla4intioctl_hba_reset(ha, ioctl);
		break;

	case INT_CC_COPY_FW_FLASH:
		status = qla4intioctl_copy_fw_flash(ha, ioctl);
		break;

	default:
		QL4PRINT(QLP1|QLP4, printk("qla4000_ioctl: unsupported command code (%X)\n", cmd));
		ioctl->Status = EXT_STATUS_UNSUPPORTED_SUBCODE;
		status = (-EINVAL);
		up(&ha->ioctl_sem);
		goto exit_qla4000_ioctl;
	}

	/*
	 * Copy the updated ioctl structure back to the user
	 */
	copy_to_user(arg, (void *)ioctl, sizeof(EXT_IOCTL_ISCSI));

	up(&ha->ioctl_sem);

exit_qla4000_ioctl:

	LEAVE("qla4000_ioctl");
	return(status);
}

/*
 * Overrides for Emacs so that we almost follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 2
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -2
 * c-argdecl-indent: 2
 * c-label-offset: -2
 * c-continued-statement-offset: 2
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */
